
################################################################################
#'
#' Reads a TAS-created signal.txt file.
#'
#' Reads a human-readable signal.txt file created by the Tiling Array Software
#' (TAS) from Affymetrix. The signal files comprise information on the genomic
#' map entry (e.g. chromosome name), probe center positions and probe
#' intensities. The intensities are assumed to be in \code{log2} scale. The
#' method generates a \code{list} for each genomic map entry containing the
#' center position and intensity of the corresponding probes.
#'
#' @param  filename  Filename of human-readable signal.txt file
#'         (as a \code{character}) created by the Tiling Array Software (TAS).
#' @param  minhits Minimal number of hits in BPMAP entry to be considered for
#'         the further analysis. Due to historical reasons there are several
#'         entries in the BPMAP file with only around thousand probes assigned
#'         that might overlap with the larger entries or with entries on other
#'         tiling arrays.
#' @param  verbose Indicates whether information on progress are printed.
#' @return Returns a \code{list} for each genomic map entry containing the
#'         center position and intensity of all corresponding probes.
#' @keywords internal
#'
TileReadSignalHelper <- function (filename,
                                  minhits,
                                  verbose) {
################################################################################
  stopifnot(!missing(filename) && !missing(minhits) && !missing(verbose))

  l <- readLines(con=filename)
  n <- length(l)
  start <- -1
  bar <- list()
  chr <- character()
  gmap <- character()
  for(i in 1:n){
    line <- l[i]
    ##if (grepl("# Genomic Map\t", line)){
    ## line starts with #
    ## (1) read next chromosome name
    ## (2) read genomic map (reference to bpmap file
    if (substr(line, 1, 1) == "#"){
      if (grepl("# Genomic Map\t", line)){
        gmap <- gsub("# Genomic Map\t", "", line)
      }

      if (grepl("# Name\t", line)){
        if (start != -1){
          list <- scan(file=filename, skip=start, nmax=i-start+1,
                       blank.lines.skip=TRUE, sep = "\t", comment.char = "#",
                       what=list(centerpos=integer(), intensity=numeric()),
                       quiet=T)
          if (length(list$centerpos) >= minhits){
            list$name = chr
            list$gmap_name = gmap
            bar <- c(bar, list(list))
          }
          else {
            bar <- c(bar, list(list(centerpos=integer(), gmap_name=gmap)))
          }
        }
        start <- i
        chr <- gsub("# Name\t", "", line)
      }
    }
  }
  list <- scan(file=filename, skip=start, nmax=n-start+1,
               blank.lines.skip=TRUE, sep = "\t", comment.char = "#",
               what=list(centerpos=integer(), intensity=numeric()),
               quiet=T)
  if (length(list$centerpos) >= minhits){
    list$name = chr
    list$gmap_name = gmap
    bar <- c(bar, list(list))
  }
  else {
    bar <- c(bar, list(list(centerpos=integer(), gmap_name=gmap)))
  }
  l <- NULL
  return(bar)
}

################################################################################
#'
#' Checks parameter for exported function \code{\link{TileReadSignal}}.
#'
#' Checks parameter for exported function \code{\link{TileReadCel}} whether they
#' are missing, have a incorrect data type, or are obviously not correct, e.g.,
#' filename parameter refers to non-existing file.
#'
#' @param  signal.filename  Filename as a \code{character} to existing file.
#' @param  bpmap.filename Filename as a \code{character} to existing file.
#' @param  minhits An \code{integer} value.
#' @param  group A \code{character} value.
#' @param  gc A \code{logical} value.
#' @param  matchscore A \code{logical} value.
#' @param  verbose A \code{logical} value.
#' @return None.
#' @keywords internal
#'
TileReadSignalCheck <- function(signal.filename,
                                bpmap.filename,
                                minhits,
                                group,
                                gc,
                                matchscore,
                                verbose){
################################################################################
  if (missing(signal.filename) || !is.character(signal.filename) ||
      length(signal.filename) != 1){
    stop("The parameter 'signal.filename' is missing or invalid.")
  }
  else {
    if (!file.exists(signal.filename)){
      stop("TAS-created signal.txt file not found.")
    }
  }
  if (missing(bpmap.filename) || !is.character(bpmap.filename) ||
      length(bpmap.filename) != 1){
    stop("The parameter 'bpmap.filename' is missing or invalid.")
  }
  else {
    if (!file.exists(bpmap.filename)){
      stop("Affymetrix BPMAP file not found.")
    }
  }
  if (missing(minhits) || !(is.integer(minhits) || is.numeric(minhits)) ||
      round(minhits) != minhits || length(minhits) != 1){
    stop("The parameter 'minhits' is missing or invalid.")
  }
  if (missing(group) || !is.character(group) || length(group) != 1){
    stop("The parameter 'group' is missing or invalid.")
  }
  if (missing(gc) || !is.logical(gc) || length(gc) != 1){
    stop("The parameter 'gc' is missing or invalid.")
  }
  if (missing(matchscore) || !is.logical(matchscore) ||
      length(matchscore) != 1){
    stop("The parameter 'matchscore' is missing or invalid.")
  }
  if (missing(verbose) || !is.logical(verbose) || length(verbose) != 1){
    stop("The parameter 'verbose' is missing or invalid.")
  }
}

################################################################################
#'
#' Reads tiling array data from an Affymetrix BPMAP file and a TAS-created
#' signal.txt file.
#'
#' Reads tiling array data in terms of an Affymetrix BPMAP (binary probe
#' mapping) file and the human-readable signal.txt files that are created by the
#' Tiling Array Software (TAS) from Affymetrix. The signal files comprise
#' information on the genomic map entry (e.g. chromosome name), probe center
#' positions and probe intensities. The intensities are assumed to be in
#' \code{log2} scale. The method generates a \code{data.frame} comprising all
#' required data on probes that are necessary for the subsequent shuffling
#' analysis.
#'
#' @param  signal.filename  Filename of human-readable signal.txt file
#'         (as a \code{character}) created by the Tiling Array Software (TAS).
#' @param  bpmap.filename Filename of Affymetrix binary probe mapping (BPMAP)
#'         file (as a \code{character}), which is a binary file containing
#'         information on the location of each probe in the reference sequence.
#'         Moreover, it stores the probe sequences that are necessary to
#'         calculate the GC content.
#' @param  minhits Minimal number of hits in BPMAP entry to be considered for
#'         the further analysis. Due to historical reasons there are several
#'         entries in the BPMAP file with only around thousand probes assigned
#'         that might overlap with the larger entries or with entries on other
#'         tiling arrays. In case of Affy tiling array 1.0R, a value of 8000
#'         is recommended.
#' @param  group A group name as the organism abbreviation in order to consider
#'         only these entries in the BPMAP file and hence disregard entries such
#'         as TIGR, Affymetrix, or bacterial controls.
#' @param  gc Indicates whether GC content of probe sequences will be
#'         calculated. It is defined as fraction of both Gs and Cs in the
#'         probe sequence.
#' @param  matchscore Indicates whether match score will be read. The match
#'         score is defined as number of perfect matches of the probe sequence
#'         per megabase of the genomic sequence. This information needs to be
#'         set accordingly in the BPMAP file. Otherwise, all probes contain only
#'         the default value.
#' @param  verbose Indicates whether information on progress are printed.
#' @return Returns a \code{data.frame} containing information on all probes,
#'         i.e., the full name of the reference sequence (organism abbreviation
#'         and chromosome name), the chromosome name, the probe center position,
#'         the length of the probe sequence, the GC content of the probe
#'         sequence, the match score (if \code{matchscore} is enabled), and
#'         the \code{log2} probe intensity or the probe \code{log2}-fold change.
#' @note   This method is deprecated since it was specifically designed
#'         for the Affymetrix Human tiling array 1.0R platform and the files
#'         created by TAS. Please consider using \code{\link{TileReadCel}} or
#'         \code{\link{TileReadCustom}} instead.
#' @import affxparser
#' @include TileHelper.R
#' @export
#'
TileReadSignal <- function(signal.filename,
                           bpmap.filename,
                           minhits = 8000,
                           group = "Hs",
                           gc = TRUE,
                           matchscore = FALSE,
                           verbose = FALSE){
################################################################################

  ## check given parameters
  TileReadSignalCheck(signal.filename=signal.filename,
                      bpmap.filename=bpmap.filename, minhits=minhits,
                      group=group, gc=gc, matchscore=matchscore,
                      verbose=verbose)

  if (verbose){
    TileMsg("Reading tiling array data started with ",
            "signal.filename=", TileConvPath(filename=signal.filename),
            ", bpmap.filename=", TileConvPath(filename=bpmap.filename),
            ", minhits=", minhits, ", group=", group, ", gc=", gc,
            ", matchscore=", matchscore)
  }

  signal <- TileReadSignalHelper(filename=signal.filename,
                                 minhits=minhits,
                                 verbose=verbose)
  chip <- NULL

  if (basename(signal[[1]]$gmap_name) != basename(bpmap.filename)){
    TileMsg(file=stderr(), "Warning: BPMAP entry '",
            TileConvPath(filename=signal[[1]]$gmap_name),
            "' in signal file differs from given BPMAP filename '",
            TileConvPath(filename=bpmap.filename), "'.")
  }
  bpmap.filename <- TileConvPath(filename=bpmap.filename)
  bpmap.info <- readBpmapSeqinfo(bpmap.filename, verbose=0)
  idx <- setdiff(1:length(bpmap.info),
                 as.integer(which(unlist(lapply(bpmap.info,
                 function(x){grepl("AffxCtrl", x$groupname)})) == TRUE)))
  for (i in 1:length(signal)){
    if (length(signal[[i]]$centerpos) == 0){
      next
    }
    bpmap.idx <- NULL
    ## new version (entry number of bar files is equal to entry number
    ## in bpmap file by omitting the AffxCtrl entries (very unsafe!!!)
    if (i <= length(bpmap.info) &&
        bpmap.info[[idx[i]]]$groupname == group &&
        bpmap.info[[idx[i]]]$name == signal[[i]]$name){
      bpmap.idx <- c(idx[i])
      bpmap <- readBpmap(bpmap.filename, seqIndices=bpmap.idx,
                         readCenterPos=TRUE, readStartPos=FALSE,
                         readPMXY=FALSE, readMMXY=FALSE, readProbeLength=TRUE,
                         readMatchScore=matchscore,
                         verbose=0)
      for (j in 1:length(bpmap.idx)){
        if (verbose){
          TileMsg("Processing sequence object ", bpmap[[j]]$seqInfo$fullname)
        }
        chr <- data.frame(fullname=as.character(bpmap[[j]]$seqInfo$fullname),
                          name=as.character(bpmap[[j]]$seqInfo$name),
                          centerpos=as.integer(bpmap[[j]]$centerpos),
                          probelength=as.integer(bpmap[[j]]$probelength))
        if (matchscore){
          chr$matchscore <- as.numeric(bpmap[[j]]$matchscore)
        }

        ## merge Signal and BPMAP file to assign intensities
        chr$pos <- 1:nrow(chr)
        if (verbose){
          TileMsg("Assigning intensities from signal file to probes")
        }
        tmp <- merge(chr, signal[[i]], by="centerpos", all.x=TRUE, sort=FALSE)
        chr$intensity <- as.numeric(tmp$intensity[order(tmp$pos)])
        chr$pos <- NULL
        stopifnot(all(!is.na(chr$intensity)))

        ## calculate GC content
        if (gc){
          if (verbose){
            TileMsg("Calculating GC content of probes")
          }
          chr$gc <- as.numeric(TileCalcGC(probeseq=bpmap[[j]]$probeseq))
        }
        else {
          chr$gc <- 0
        }
        chip <- rbind(chip, chr)
        rm(chr, tmp)
      }
      rm(bpmap, bpmap.idx)
    }
  }
  if (is.null(chip) || nrow(chip) == 0){
    stop("No data matching conditions in current signal file.")
  }
  rm(signal)
  rm(bpmap.info)
  return(chip)
}

################################################################################
#'
#' Checks parameter for exported function \code{\link{TileReadCel}}.
#'
#' Checks parameter for exported function \code{\link{TileReadCel}} whether they
#' are missing, have a incorrect data type, or are obviously not correct, e.g.,
#' filename parameter refers to non-existing file.
#'
#' @param  cel.filename A \code{vector} of one or more filenames as a
#'         \code{character} to existing files.
#' @param  cel2.filename A \code{vector} of one or more filename as a
#'         \code{character} to existing files or set to \code{NULL}.
#' @param  celinc.filename A \code{vector} of filenames or one or more
#'         filenames as \code{character} to existing files or set to
#'         \code{NULL}.
#' @param  bpmap.filename Filename as a \code{character} to existing file.
#' @param  minhits An \code{integer} value.
#' @param  group A \code{character} value.
#' @param  pmonly A \code{logical} value that cannot be enabled if \code{mmonly}
#'         is \code{TRUE}.
#' @param  mmonly A \code{logical} value that cannot be enabled if \code{pmonly}
#'         is \code{TRUE}.
#' @param  normalize A \code{logical} value.
#' @param  gc A \code{logical} value.
#' @param  matchscore A \code{logical} value.
#' @param  verbose A \code{logical} value.
#' @return None.
#' @keywords internal
#'
TileReadCelCheck <- function(cel.filename,
                             cel2.filename,
                             celinc.filename,
                             bpmap.filename,
                             minhits,
                             group,
                             pmonly,
                             mmonly,
                             normalize,
                             gc,
                             matchscore,
                             verbose){
################################################################################
  if (missing(cel.filename) || any(!is.character(cel.filename))){
    stop("The parameter 'cel.filename' is missing or invalid.")
  }
  else {
    if (any(!file.exists(cel.filename))){
      stop("Affymetrix CEL files in first condition not found.")
    }
  }
  if (missing(cel2.filename) ||
      (!is.null(cel2.filename) && any(!is.character(cel2.filename)))){
    stop("The parameter 'cel2.filename' is missing or invalid.")
  }
  else {
    if (!is.null(cel2.filename) && any(!file.exists(cel2.filename))){
      stop("Affymetrix CEL files in second condition not found.")
    }
  }
  if (missing(celinc.filename) ||
      (!is.null(celinc.filename) && any(!is.character(celinc.filename)))){
    stop("The parameter 'celinc.filename' is missing or invalid.")
  }
  else {
    if (!is.null(celinc.filename) && any(!file.exists(celinc.filename))){
      stop("Additional Affymetrix CEL files for normalization not found.")
    }
  }
  if (missing(bpmap.filename) || !is.character(bpmap.filename) ||
      length(bpmap.filename) != 1){
    stop("The parameter 'bpmap.filename' is missing or invalid.")
  }
  else {
    if (!file.exists(bpmap.filename)){
      stop("Affymetrix BPMAP file not found.")
    }
  }
  if (missing(minhits) || !(is.integer(minhits) || is.numeric(minhits)) ||
      round(minhits) != minhits || length(minhits) != 1){
    stop("The parameter 'minhits' is missing or invalid.")
  }
  if (missing(group) || !is.character(group) || length(group) != 1){
    stop("The parameter 'group' is missing or invalid.")
  }
  if (missing(pmonly) || !is.logical(pmonly) || length(pmonly) != 1){
    stop("The parameter 'pmonly' is missing or invalid.")
  }
  if (missing(mmonly) || !is.logical(mmonly) || length(mmonly) != 1){
    stop("The parameter 'mmonly' is missing or invalid.")
  }
  if (missing(normalize) || !is.logical(normalize) || length(normalize) != 1){
    stop("The parameter 'normalize' is missing or invalid.")
  }
  if (missing(gc) || !is.logical(gc) || length(gc) != 1){
    stop("The parameter 'gc' is missing or invalid.")
  }
  if (missing(matchscore) || !is.logical(matchscore) ||
      length(matchscore) != 1){
    stop("The parameter 'matchscore' is missing or invalid.")
  }
  if (missing(verbose) || !is.logical(verbose) || length(verbose) != 1){
    stop("The parameter 'verbose' is missing or invalid.")
  }

  ## additional checks
  if (pmonly && mmonly){
    stop("The parameters 'pmonly' and 'mmonly' are mutually exclusive.")
  }
  if (!is.null(celinc.filename) && !normalize){
    stop("Additional CEL files given but normalization is disabled.")
  }
}

################################################################################
#'
#' Reads tiling array data from an Affymetrix BPMAP file and CEL file(s).
#'
#' Reads tiling array data in terms of an Affymetrix BPMAP (binary probe
#' mapping) file and the Affymetrix CEL file that are created by the Affymetrix
#' pipeline. The CEL file stores the results of the intensity calculations on
#' the pixel values of the DAT file. This includes among others an intensity
#' value, standard deviation of the intensity, and the coordinates of each
#' PM/MM probe on the tiling array. These coordinate information can be merged
#' with the Affymetrix BPMAP file that also comprises further information on the
#' probes. This includes the genomic location of each probe in terms of the
#' reference sequence and the probe center position, probe sequence for the GC
#' calculation, and the corresponding probe sequence length. The method
#' generates a \code{data.frame} comprising all required data on probes that
#' are necessary for the subsequent shuffling analysis.
#'
#' @param  cel.filename A \code{vector} of one or more filenames of Affymetrix
#'         CEL files (as \code{character}) that contain the probe intensities in
#'         the first cellular condition. Note that replicates are simply defined
#'         by more than one filename. In case \code{cel2.filename} is set to
#'         \code{NULL}, the median probe \code{log2}-intensity among the given
#'         replicates (\code{cel.filename}) is reported as estimate of the probe
#'         intensity. On the other hand, if \code{cel2.filename} is given, the
#'         median of all possible pairwise \code{log2}-fold changes between CEL
#'         files from the first cellular condition (\code{cel.filename}) and
#'         ones from the second cellular condition (\code{cel2.filename}) of a
#'         probe will be reported as estimate of the probe intensity change.
#' @param  cel2.filename A \code{vector} of one or more filenames of Affymetrix
#'         CEL files (as \code{character}) that contain the probe intensities in
#'         the second cellular condition. Note that replicates are simply
#'         defined by more than one filename. This parameter is only required
#'         in case of differential expression analysis where the median of all
#'         possible pairwise \code{log2}-fold changes between CEL files from
#'         the first cellular condition (\code{cel.filename}) and ones from the
#'         second cellular condition (\code{cel2.filename}) of a probe will be
#'         reported as estimate for the probe intensity change. Otherwise, the
#'         parameter must be set to \code{NULL}.
#' @param  celinc.filename A \code{vector} of one or more filenames of
#'         Affymetrix CEL files (as \code{character}) containing probe
#'         intensities that should be included in the normalization. This may
#'         be desirable in case tiling array data of more than two different
#'         cellular states is available and multiple transitions between them
#'         are being analyzed. In the analysis of any of these transitions,
#'         the CEL files corresponding to the remaining cellular states, i.e.,
#'         those not given by \code{cel.filename} or \code{cel2.filename},
#'         may be defined as \code{celinc.filename}. Hence, the full-quantile
#'         normalization is always done on the entire set of avaiable intensity
#'         data and the \code{log2}-fold changes among different analyzed
#'         transitions are comparable and not biased by the normalization
#'         procedure. This parameter is futile if \code{normalization} is
#'         disabled.
#' @param  bpmap.filename Filename of Affymetrix binary probe mapping (BPMAP)
#'         file (as a \code{character}), which is a binary file containing
#'         information on the location of each probe in the reference sequence.
#'         Moreover, it stores the probe sequences that are necessary to
#'         calculate the GC content.
#' @param  minhits Minimal number of hits in BPMAP entry to be considered for
#'         the further analysis. Due to historical reasons there are several
#'         entries in the BPMAP file with only around thousand probes assigned
#'         that might overlap with the larger entries or with entries on other
#'         tiling arrays. In case of Affy tiling array 1.0R, a value of 8000
#'         is recommended.
#' @param  group A group name as the organism abbreviation in order to consider
#'         only these entries in the BPMAP file and hence disregard entries such
#'         as TIGR, Affymetrix, or bacterial controls.
#' @param  pmonly Indicates whether only intensities of perfect match (PM)
#'         probes on the Affymetrix tiling array are incorporated in the probe
#'         intensity estimation. If neither \code{pmonly} nor \code{mmonly} is
#'         set to \code{TRUE}, the specific hybridization effect of a probe is
#'         estimated by taking \code{PM-MM}. Due to the issue of higher
#'         intensities of mismatch (MM) probes compared to PM probes, it is
#'         recommended to enable this parameter.
#' @param  mmonly Indicates whether only intensities of mismatch (MM) probes
#'         on the Affymetrix tiling array are incorporated in the probe
#'         intensity estimation. If neither \code{pmonly} nor \code{mmonly} is
#'         set to \code{TRUE}, the specific hybridaztion effect of a probe is
#'         estimated by taking \code{PM-MM}. This option is mutually exclusive
#'         with the \code{pmonly} parameter and is only recommended for
#'         investigating the behaviour of mismatch probes on a tiling array
#'         but not in common (differential) expression analyses.
#' @param  normalize Indicates whether the probe intensities of the given CEL
#'         files in \code{cel.filename}, \code{cel2.filename}, and
#'         \code{celinc.filename} are normalized by use of full-quantile
#'         normalization. The normalization is recommended if replicates are
#'         available or a differential analysis is executed and, hence, the
#'         transition between cellular states is analyzed. Note that both PM
#'         and MM probes are included in the normalization regardless of the
#'         \code{pmonly} or \code{mmonly} parameter.
#' @param  gc Indicates whether GC content of probe sequences will be
#'         calculated. It is defined as fraction of both Gs and Cs in the
#'         probe sequence.
#' @param  matchscore Indicates whether match score will be read. The match
#'         score is defined as number of perfect matches of the probe sequence
#'         per megabase of the genomic sequence. This information needs to be
#'         set accordingly in the BPMAP file. Otherwise, all probes contain only
#'         the default value.
#' @param  verbose Indicates whether information on progress are printed.
#' @return Returns a \code{data.frame} containing information on all probes,
#'         i.e., the full name of the reference sequence (organism abbreviation
#'         and chromosome name), the chromosome name, the probe center position,
#'         the length of the probe sequence, the GC content of the probe
#'         sequence, the match score (if \code{matchscore} is enabled), and
#'         the \code{log2} probe intensity or the probe \code{log2}-fold change.
#' @import affxparser
#' @import Biobase
#' @include TileHelper.R
#' @export
#' @examples
#' ## Note that the following example only executes if the external data
#' ## of the Starr R package is available which includes an artificial
#' ## Affymetrix BPMAP file and corresponding CEL files.
#' path <- system.file("extdata", package = "Starr")
#' if (path != ""){
#'   ## define Affymetrix BPMAP file for probe mapping
#'   bpmap.filename <- file.path(path, "Scerevisiae_tlg_chr1.bpmap")
#'   ## define Affymetrix CEL files
#'   ## here: file of control experiment (wt)
#'   wt.filename <- file.path(path, "wt_IP_chr1.cel")
#'   stopifnot(file.exists(bpmap.filename) && file.exists(wt.filename))
#'
#'   ## read CEL file and return data.frame
#'   ## with information such as genomic localization
#'   ## of probes, GC content of probe sequences,
#'   ## and probe intensities.
#'   ## Note that group is '' (blank) for old Affy chr21/22arrays
#'   ## while it is "Hs" for other Human, "Mm" for Mouse, or "Dm"
#'   ## for Drosophila tiling array platforms.
#'   wt.cel <- TileReadCel(cel.filename=wt.filename,
#'                         bpmap.filename=bpmap.filename,
#'                         group="", gc=TRUE, verbose=FALSE)
#'
#'   ## getting an overview on the reported data.frame
#'   str(wt.cel)
#'
#'   ## investigating data
#'   ## e.g. plot density of intensities
#'   pdf(file="wt_int_density.pdf")
#'   plot(density(wt.cel$intensity), main="", xlab="Intensity")
#'   dev.off()
#'   ## or GC bias with three GC content bins
#'   pdf(file="wt_gc_bias.pdf")
#'   boxplot(wt.cel$intensity ~ cut(wt.cel$gc,
#'           breaks=c(0,0.36,0.52,Inf),right=FALSE),
#'           xlab="GC content", ylab="Intensity")
#'   dev.off()
#'   ## cleanup
#'   rm(bpmap.filename, wt.filename, wt.cel)
#' }
#' rm(path)
#'
TileReadCel <- function(cel.filename,
                        cel2.filename = NULL,
                        celinc.filename = NULL,
                        bpmap.filename,
                        minhits = 8000,
                        group = "Hs",
                        pmonly = TRUE,
                        mmonly = FALSE,
                        normalize = TRUE,
                        gc = TRUE,
                        matchscore = FALSE,
                        verbose = FALSE){
################################################################################
  bpmap <- NULL; cel <- NULL; cel2 <- NULL; celinc <- NULL

  ## check given parameters
  TileReadCelCheck(cel.filename=cel.filename, cel2.filename=cel2.filename,
                   celinc.filename=celinc.filename,
                   bpmap.filename=bpmap.filename,
                   minhits=minhits, group=group, pmonly=pmonly, mmonly=mmonly,
                   normalize=normalize, gc=gc, matchscore=matchscore,
                   verbose=verbose)

  if (verbose){
    TileMsg("Reading tiling array data started with ",
            paste("cel.filename=", TileConvPath(filename=cel.filename),
                  ", ", sep="", collapse=""),
            paste("cel2.filename=", TileConvPath(filename=cel2.filename),
                  ", ", sep="", collapse=""),
            paste("celinc.filename=", TileConvPath(filename=celinc.filename),
                  ", ", sep="", collapse=""),
            "bpmap.filename=", TileConvPath(filename=bpmap.filename),
            ", minhits=", minhits, ", group=", group, ", pmonly=", pmonly,
            ", mmonly=", mmonly, ", normalize=", normalize, ", gc=", gc,
            ", matchscore=", matchscore)
  }

  ## read bpmap and cel file
  bpmap <- readBpmap(filename=TileConvPath(filename=bpmap.filename),
                     readCenterPos = TRUE, readStartPos = FALSE,
                     readPMXY = !mmonly, readMMXY = !pmonly,
                     readProbeLength = TRUE, readMatchScore = matchscore,
                     verbose = 0)
  if (is.null(bpmap)){
    stop("Reading BPMAP file was not successful.")
  }
  for (i in cel.filename){
    cel <- c(cel, list(readCel(filename=i, readHeader = TRUE, readXY = TRUE,
                               readIntensities = TRUE, readOutliers = FALSE,
                               readMasked = FALSE, verbose = 0)))
  }

  ## warning if chiptype of bpmap and cel does not fit
  ##if (!grepl(cel$header$chiptype, basename(bpmap.filename))){
  ##  TileMsg(file=stderr(), "Warning: Chiptype '", cel$header$chiptype,
  ##          "' of CEL file differs from BPMAP filename '",
  ##          TileConvPath(filename=bpmap.filename), "'.")
  ##}

  if (!is.null(cel2.filename)){
    for (i in cel2.filename){
      cel2 <- c(cel2, list(readCel(filename=i, readHeader = TRUE, readXY = TRUE,
                                   readIntensities = TRUE, readOutliers = FALSE,
                                   readMasked = FALSE, verbose = 0)))
    }
    ## warning if chiptype of bpmap and cel does not fit
    ##if (!grepl(cel2$header$chiptype, basename(bpmap.filename))){
    ##  TileMsg(file=stderr(), "Warning: Chiptype '", cel2$header$chiptype,
    ##          "' of second CEL file differs from BPMAP filename '",
    ##          TileConvPath(filename=bpmap.filename), "'.")
    ##}
  }
  if (!is.null(celinc.filename)){
    for (i in celinc.filename){
      celinc <- c(celinc, list(readCel(filename=i, readHeader = TRUE,
                                       readXY = FALSE, readIntensities = TRUE,
                                       readOutliers = FALSE, readMasked = FALSE,
                                       verbose = 0)))
    }
  }

  ## normalization
  if (normalize){
    if (verbose){
      TileMsg("Quantile normalizing intensities among states")
    }
    ## collect data
    data <- NULL
    for (i in c(cel, cel2, celinc)){
      if (!is.null(data) && length(data[,1]) !=
          length(i$intensities)){
        stop("Different number of probes in given CEL files")
      }
      data <- cbind(data, i$intensities)
    }

    ## quantile normalize
    data <- TileQuantileNorm(int=data, overwrite=FALSE)

    ## write back normalized values
    for (i in 1:length(cel)){
      cel[[i]]$intensities <- data[,i]
    }
    if (!is.null(cel2.filename)){
      for (i in 1:length(cel2)){
        cel2[[i]]$intensities <- data[,i+length(cel)]
      }
    }
    rm(data)
  }

  chip <- NULL
  ## iterate through all bpmap entries
  for (i in 1:length(bpmap)){
    if (bpmap[[i]]$seqInfo$groupname == group &&
        bpmap[[i]]$seqInfo$numberOfHits >= minhits){
      if (verbose){
        TileMsg("Processing sequence object ", bpmap[[i]]$seqInfo$fullname)
      }
      ## initialize data.frame
      chr <- data.frame(fullname=as.character(rep(bpmap[[i]]$seqInfo$fullname,
                          bpmap[[i]]$seqInfo$numberOfHits)),
                        name=as.character(rep(bpmap[[i]]$seqInfo$name,
                          bpmap[[i]]$seqInfo$numberOfHits)),
                        centerpos=as.integer(bpmap[[i]]$centerpos),
                        probelength=as.integer(bpmap[[i]]$probelength))

      if (matchscore){
        chr$matchscore <- as.numeric(bpmap[[i]]$matchscore)
      }

      if (!mmonly){
        chr$pmx=bpmap[[i]]$pmx
        chr$pmy=bpmap[[i]]$pmy
      }

      if (!pmonly){
        chr$mmx=bpmap[[i]]$mmx
        chr$mmy=bpmap[[i]]$mmy
      }
      ## calculate GC content
      if (gc){
        if (verbose){
          TileMsg("Calculating GC content of probes")
        }
        chr$gc <- as.numeric(TileCalcGC(probeseq=bpmap[[i]]$probeseq))
      }
      else {
        chr$gc <- 0
      }
      chr <- chr[order(chr$name, chr$centerpos, chr$probelength),]
      chr$idx <- 1:nrow(chr)

      ## extract intensities, generates matrix
      ## with columns for each cel and cel2 file
      ## and rows for each probe (sorted by chr$idx)
      if (verbose){
        TileMsg("Assigning intensities to probes")
      }
      pmint <- NULL
      mmint <- NULL
      for (j in c(cel, cel2)){
        if (!mmonly){
          merged <- merge(chr[c("pmx", "pmy", "idx")],
                          j[c("intensities", "x", "y")],
                          by.x=c("pmx", "pmy"), by.y=c("x", "y"),
                          all.x=TRUE)
          stopifnot(all(!is.na(merged)))
          pmint <- cbind(pmint, merged[order(merged$idx),"intensities"])
          rm(merged)
        }
        if (!pmonly){
          merged <- merge(chr[c("mmx", "mmy", "idx")],
                          j[c("intensities", "x", "y")],
                          by.x=c("mmx", "mmy"), by.y=c("x", "y"),
                          all.x=TRUE)
          stopifnot(all(!is.na(merged)))
          mmint <- cbind(mmint, merged[order(merged$idx),"intensities"])
          rm(merged)
        }
      }

      ## calculate log2 scores
      ## and eventually PM-MM
      if (pmonly){
        int <- log2(pmax(pmint, 1))
      }
      else {
        if (mmonly){
          int <- log2(pmax(mmint, 1))
        }
        else {
          int <- log2(pmax(pmint-mmint, 1))
        }
      }
      ## one-sample:
      ## simply take rowMedians
      ## over replicates
      if (is.null(cel2.filename)){
        if (ncol(int) > 1){
          chr$intensity <- as.numeric(rowMedians(imat=int))
        }
        else {
          chr$intensity <- as.numeric(int[,1])
        }
      }
      ## two-samples:
      ## take paired median difference
      ## (median among all combinations
      ## of intensity differences in terms
      ## of cel2-cel)
      else {
        intcomb <- NULL
        for (j in 1:length(cel)){
          for (k in 1:length(cel2)){
            ## INFO: int stores cel intensities from 1 to length(cel)
            ## and cel2 intensities from length(cel)+1 to
            ## length(cel) + length(cel2)
            intcomb <- cbind(intcomb, int[,k+length(cel)]-int[,j])
          }
        }
        if (ncol(int) > 1){
          chr$intensity <- as.numeric(rowMedians(imat=intcomb))
        }
        else {
          chr$intensity <- as.numeric(intcomb[,1])
        }
        rm(intcomb)
      }

      # cleanup
      chr$idx <- NULL
      if (!mmonly){
        rm(pmint)
        chr$pmx <- NULL; chr$pmy <- NULL
      }
      if (!pmonly){
        rm(mmint)
        chr$mmx <- NULL; chr$mmy <- NULL
      }
      chip <- rbind(chip, chr)
      rm(int, chr)
    }
  }
  rm(bpmap, cel)
  if (!is.null(cel2.filename)){
    rm(cel2)
  }
  if (!is.null(celinc.filename)){
    rm(celinc)
  }
  ##browser()
  return(chip)
}

################################################################################
#'
#' Checks parameter for exported function \code{\link{TileReadCustom}}.
#'
#' Checks parameter for exported function \code{\link{TileReadCustom}}
#' whether they are missing, have a incorrect data type, or are obviously
#' not correct, e.g., filename parameter refers to non-existing file.
#'
#' @param  custom.filename A \code{vector} of one or more filenames as a
#'         \code{character} to existing files.
#' @param  custom2.filename A \code{vector} of one or more filename as a
#'         \code{character} to existing files or set to \code{NULL}.
#' @param  custominc.filename A \code{vector} of filenames or one or more
#'         filenames as \code{character} to existing files or set to
#'         \code{NULL}.
#' @param  pmonly A \code{logical} value that cannot be enabled if \code{mmonly}
#'         is \code{TRUE}.
#' @param  mmonly A \code{logical} value that cannot be enabled if \code{pmonly}
#'         is \code{TRUE}.
#' @param  normalize A \code{logical} value.
#' @param  gc A \code{logical} value.
#' @param  verbose A \code{logical} value.
#' @return None.
#' @keywords internal
#'
TileReadCustomCheck <- function(custom.filename,
                                custom2.filename,
                                custominc.filename,
                                pmonly,
                                mmonly,
                                normalize,
                                gc,
                                verbose){
################################################################################
    if (missing(custom.filename) || any(!is.character(custom.filename))){
        stop("The parameter 'custom.filename' is missing or invalid.")
    }
    else {
        if (any(!file.exists(custom.filename))){
            stop("Custom-formatted files in first condition not found.")
        }
    }
    if (missing(custom2.filename) ||
        (!is.null(custom2.filename) && any(!is.character(custom2.filename)))){
        stop("The parameter 'custom2.filename' is missing or invalid.")
    }
    else {
        if (!is.null(custom2.filename) && any(!file.exists(custom2.filename))){
            stop("Custom-formatted files in second condition not found.")
        }
    }
    if (missing(custominc.filename) ||
        (!is.null(custominc.filename) &&
         any(!is.character(custominc.filename)))){
        stop("The parameter 'custominc.filename' is missing or invalid")
    }
    else {
        if (!is.null(custominc.filename) &&
            any(!file.exists(custominc.filename))){
            stop("Additional custom-formatted files for ",
                 "normalization not found.")
        }
    }
    if (missing(pmonly) || !is.logical(pmonly) || length(pmonly) != 1){
        stop("The parameter 'pmonly' is missing or invalid.")
    }
    if (missing(mmonly) || !is.logical(mmonly) || length(mmonly) != 1){
        stop("The parameter 'mmonly' is missing or invalid.")
    }
    if (missing(normalize) || !is.logical(normalize) || length(normalize) != 1){
        stop("The parameter 'normalize' is missing or invalid.")
    }
    if (missing(gc) || !is.logical(gc) || length(gc) != 1){
        stop("The parameter 'gc' is missing or invalid.")
    }
    if (missing(verbose) || !is.logical(verbose) || length(verbose) != 1){
        stop("The parameter 'verbose' is missing or invalid.")
    }

    ## additional checks
    if (pmonly && mmonly){
        stop("The parameter 'pmonly' and 'mmonly' are mutually exclusive.")
    }
    if (!is.null(custominc.filename) && !normalize){
        stop("Additional custom-formatted files given ",
             "but normalization disabled.")
    }
}

################################################################################
#'
#' Checks the data for exported function \code{\link{TileReadCustom}}.
#'
#' Checks the data for exported function \code{\link{TileReadCustom}}
#' in terms of validity, e.g., uniqueness of probe identifiers and
#' obvious errors in probe coordinates, and consistency to previously
#' read data, e.g., same probe identifiers, probe coordinates, and probe
#' sequences.
#'
#' @param  refdata A \code{data.frame} of reference data that was already
#'         read and which is used for consistency checks with current data.
#'         In case of the first read dataset, it is set to \code{NULL} and
#'         consistency checks are not performed.
#' @param  data A \code{data.frame} that is checked for validity and
#'         consistency.
#' @return None.
#' @keywords internal
#'
TileReadCustomCheckData <- function(refdata, data){
################################################################################
    ## check uniqueness of probeid
    if (any(duplicated(data$probeid))){
        stop("Probe identifier is not unique.")
    }
    ## further checks
    if (any(data$start > data$end)){
        stop("Start position larger than end position of probe.")
    }
    if (any(is.na(data))){
        stop("Input data contains 'NA' values.")
    }

    if (!is.null(refdata)){
        if (any(refdata$probeid != data$probeid) ||
            any(refdata$name != data$name) ||
            any(refdata$start != data$start) ||
            any(refdata$end != data$end) ||
            any(refdata$probeseq != refdata$probeseq)){
            stop("Inconsistent probe identifiers, coordinates, or probe ",
                 "sequences among given input files.")
        }
    }
}

################################################################################
#'
#' Reads tiling array data from custom-formatted files.
#'
#' Reads tiling array data from custom-formatted files that may be created by
#' any tiling array platform. It includes probe information that is separated
#' by tabulators. Except for comment (indicated by '#' on beginning) or empty
#' line, each line must contain the following columns: probe identifier that is
#' unique within each file and consistent among the given files in terms of
#' their probe coordinates, name of the reference sequence, start and end
#' position of the probe on the reference sequence (both 0-based, as
#' \code{integer}), intensity value of PM probe (non-log scale), intensity value
#' of MM probe (on non-log scale), and probe sequence in order to calculate the
#' GC content. The data must not contain 'NA' values. Moreover, the probe
#' sequences may be set arbitrary if \code{gc} is disabled. The method generates
#' a \code{data.frame} comprising all required data on probes that are necessary
#' for the subsequent shuffling analysis.
#'
#' @param  custom.filename A \code{vector} of one or more filenames of custom-
#'         formatted files (as \code{character}) that contain the probe
#'         intensities in the first cellular condition and are formatted as
#'         described. Note that replicates are simply defined by more than
#'         one filename. In case \code{custom2.filename} is set to \code{NULL},
#'         the median probe \code{log2}-intensity among the given replicates
#'         is reported as estimate of the probe intensity. On the other hand,
#'         if \code{custom2.filename} is given, the median of all possible
#'         pairwise \code{log2}-fold changes between files from the cellular
#'         condition (\code{custom.filename}) and ones from the second cellular
#'         condition (\code{custom2.filename}) of a probe will be reported as
#'         estimate of the probe intensity change.
#' @param  custom2.filename A \code{vector} of one or more filenames of custom-
#'         formatted files (as \code{character}) that contain the probe
#'         intensities in the second cellular condition and are formatted as
#'         described. Note that replicates are simply defined by more than one
#'         one filename. This parameter is only required in case of differential
#'         expression analysis where the median of all possible pairwise
#'         \code{log2}-fold changes between files in the first cellular
#'         condition (\code{custom.filename}) and ones from the second cellular
#'         condition (\code{custom2.filename}) of a probe will be reported as
#'         estimate for the probe intensity change. Otherwise, this parameter
#'         must be set to \code{NULL}.
#' @param  custominc.filename A \code{vector} of one or more filenames of
#'         custom-formatted files (as \code{character}) containing probe
#'         intensities that should be included in the normalization. This may
#'         be desirable in case tiling array data of more than two different
#'         cellular states is available and multiple transitions between them
#'         are being analyzed. In the analysis of any of these transitions,
#'         the files corresponding to the remaining cellular states, i.e.,
#'         those not given by \code{custom.filename} or \code{custom2.filename},
#'         may be defined as \code{custominc.filename}. Hence, the full-quantile
#'         normalization is always done on the entire set of available intensity
#'         data and the \code{log2}-fold changes among different analyzed
#'         transitions are comparable and not biased by the normalization
#'         procedure. This parameter is futile if \code{normalization} is
#'         disabled.
#' @param  pmonly Indicates whether only intensities of perfect match (PM)
#'         probes on the tiling array are incorporated in the probe intensity
#'         estimation. If neither \code{pmonly} nor \code{mmonly} is set to
#'         \code{TRUE}, the specific hybridization effect of a probe is
#'         estimated by taking \code{PM-MM}.
#' @param  mmonly Indicates whether only intensities of mismatch (MM) probes
#'         are incorporated in the probe intensity estimation. If neither
#'         \code{pmonly} nor \code{mmonly} is set to \code{TRUE}, the specific
#'         hybridization effect of a probe is estimated by taking \code{PM-MM}.
#'         This option is mutually exclusive with the \code{pmonly} parameter
#'         and is only recommended for investigating the behaviour of mismatch
#'         probes but not in common (differential) expression analysis.
#' @param  normalize Indicates whether the probe intensities of the given files
#'         in \code{custom.filename}, \code{custom2.filename}, and
#'         \code{custominc.filename} are normalized by use of full-quantile
#'         normalization. The normalization is recommended if replicates are
#'         available or a differential analysis is executed and, hence, the
#'         transition between cellular states is analyzed. Note that PM and MM
#'         probe intensities are not included in the normalization if
#'         \code{mmonly} and \code{pmonly} is set to \code{TRUE}, respectively.
#' @param  gc Indicates whether GC content of probe sequences will be
#'         calculated. It is defined as fraction of both Gs and Cs in the
#'         probe sequence. The probe sequences may be set arbitrarily if
#'         \code{gc} is disabled.
#' @param  verbose Indicates whether information on progress are printed.
#' @return Returns a \code{data.frame} containing information on all probes,
#'         i.e., the name of the reference sequence, the chromosome name
#'         (here: both are equal), the probe center position, the length of
#'         the probe sequence, the GC content of the probe sequence, the match
#'         score (if \code{matchscore} is enabled), and the \code{log2} probe
#'         intensity or the probe \code{log2}-fold change.
#' @import Biobase
#' @include TileHelper.R
#' @export
#' @examples
#' ## This example requires the custom-formatted files
#' ## in the extdata folder of this package. Otherwise,
#' ## it aborts with an error.
#' path <- system.file("extdata", package = "TileShuffle")
#' stopifnot(path != "")
#' ## define filename to custom-formatted file
#' custom.filename <- file.path(path, "custom.txt")
#' stopifnot(file.exists(custom.filename))
#'
#' ## read custom-formatted file and return data.frame
#' ## with information such as genomic localization
#' ## of probes, GC content of probe sequences, and
#' ## probe intensities.
#' custom <- TileReadCustom(custom.filename=custom.filename,
#'                          pmonly=TRUE, gc=TRUE, verbose=FALSE)
#'
#' ## getting an overview on the reported data.frame
#' str(custom)
#'
#' ## investigating data
#' ## e.g. plot density of intensities
#' pdf(file="custom_int_density.pdf")
#' plot(density(custom$intensity), main="", xlab="Intensity")
#' dev.off()
#' ## or GC bias with three GC content bins
#' pdf(file="custom_gc_bias.pdf")
#' boxplot(custom$intensity ~ cut(custom$gc,
#'         breaks=c(0,0.36,0.52,Inf),right=FALSE),
#'         xlab="GC content", ylab="Intensity")
#' dev.off()
#' ## cleanup
#' rm(path, custom.filename, custom)
#'
TileReadCustom <- function(custom.filename,
                           custom2.filename = NULL,
                           custominc.filename = NULL,
                           pmonly = TRUE,
                           mmonly = FALSE,
                           normalize = TRUE,
                           gc = TRUE,
                           verbose = FALSE){
################################################################################
  custom <- NULL; custom2 <- NULL; custominc <- NULL
  col.names <- c("probeid", "name", "start", "end",
                 "pmint", "mmint", "probeseq")
  col.classes <- c("character", "character", "integer", "integer",
                   "numeric", "numeric", "character")
  
  ## check given parameter
  TileReadCustomCheck(custom.filename=custom.filename,
                      custom2.filename=custom2.filename,
                      custominc.filename=custominc.filename,
                      pmonly=pmonly, mmonly=mmonly, normalize=normalize,
                      gc=gc, verbose=verbose)

  if (verbose){
    TileMsg("Reading tiling array data started with ",
            paste("custom.filename=", TileConvPath(filename=custom.filename),
                  ", ", sep="", collapse=""),
            paste("custom2.filename=", TileConvPath(filename=
                                                    custom2.filename),
                  ", ", sep="", collapse=""),
            paste("custominc.filename=", TileConvPath(filename=
                                                      custominc.filename),
                  ", ", sep="", collapse=""),
            "pmonly=", pmonly, ", mmonly=", mmonly,
            ", normalize=", normalize, ", gc=", gc)
  }
  
  ## read custom-formatted files and check data
  ## for validity (data types, ranges, unique and
  ## equal probeids with equal probe coordinates)
  ## NOTE: data types and column number is already
  ## checked by read.table due to col.names and
  ## colClasses
  for (i in custom.filename){
    tmp <- read.table(file=i, col.names=col.names, colClasses=col.classes,
                      stringsAsFactors=F)
    tmp <- tmp[order(tmp$probeid),]
    TileReadCustomCheckData(refdata=custom[[1]], data=tmp)
    custom <- c(custom, list(tmp))
    rm(tmp)
  }
  
  if (!is.null(custom2.filename)){
    for (i in custom2.filename){
      tmp <- read.table(file=i, col.names=col.names, colClasses=col.classes,
                        stringsAsFactors=F)
      tmp <- tmp[order(tmp$probeid),]
      TileReadCustomCheckData(refdata=custom[[1]], data=tmp)
      custom2 <- c(custom2, list(tmp))
      rm(tmp)
    }
  }

  if (!is.null(custominc.filename)){
    for (i in custominc.filename){
      tmp <- read.table(file=i, col.names=col.names, colClasses=col.classes,
                        stringsAsFactors=F)
      tmp <- tmp[order(tmp$probeid),]
      TileReadCustomCheckData(refdata=custom[[1]], data=tmp)
      custominc <- c(custominc, list(tmp))
      rm(tmp)
    }
  }

  ## normalization
  if (normalize){
    if (verbose){
      TileMsg("Quantile normalizing intensities among states")
    }
    ## collect data
    ## NOTE: ignore MM intensities
    ## if pmonly and vice versa
    data <- NULL
    for (i in c(custom, custom2, custominc)){
      if (pmonly){
        data <- cbind(data, i$pmint)
      }
      else {
        if (mmonly){
          data <- cbind(data, i$mmint)
        }
        else {
          data <- cbind(data, c(i$pmint, i$mmint))
        }
      }
    }
    
    ## quantile normalize
    data <- TileQuantileNorm(int=data, overwrite=FALSE)
    
    ## write back normalized values
    j <- 1
    for (i in 1:length(custom)){
      if (pmonly){
        custom[[i]]$pmint <- data[,i]
      }
      else {
        if (mmonly){
          custom[[i]]$mmint <- data[,i]
        }
        else {
          custom[[i]]$pmint <- data[1:length(custom[[i]]$pmint),i]
          custom[[i]]$mmint <- data[length(custom[[i]]$pmint)+
                                    (1:length(custom[[i]]$mmint)),i]
        }
      }
    }
    if (!is.null(custom2.filename)){
      for (i in 1:length(custom2)){
        if (pmonly){
          custom2[[i]]$pmint <- data[,i+length(custom)]
        }
        else {
          if (mmonly){
            custom2[[i]]$mmint <- data[,i+length(custom)]
          }
          else {
            custom2[[i]]$pmint <- data[1:length(custom2[[i]]$pmint),
                                      i+length(custom)]
            custom2[[i]]$mmint <- data[length(custom2[[i]]$pmint)+
                                       (1:length(custom2[[i]]$mmint)),
                                       i+length(custom)]
          }
        }
      }
    }
    rm(data)
  }



  ## initialize data.frame
  data <- data.frame(fullname=as.character(custom[[1]]$name),
                     name=as.character(custom[[1]]$name),
                     centerpos=as.integer(TileConvStartToCenter(
                       custom[[1]]$start, custom[[1]]$end-custom[[1]]$start+1)),
                     probelength=as.integer(custom[[1]]$end-
                       custom[[1]]$start+1))

  if (gc){
    if (verbose){
      TileMsg("Calculating GC content of probes")
    }
    data$gc <- as.numeric(TileCalcGC(probeseq=custom[[1]]$probeseq))
  }
  else {
    data$gc <- 0
  }

  ## generate intensity matrix
  ## with columns for each custom
  ## and custom2 file and rows
  ## for each probe
  if (verbose){
    TileMsg("Assigning intensities to probes")
  }
  pmint <- NULL
  mmint <- NULL
  for (i in c(custom, custom2)){
    if (!mmonly){
      pmint <- cbind(pmint, i$pmint)
    }
    if(!pmonly){
      mmint <- cbind(mmint, i$mmint)
    }
  }

  ## calculate log2 scores
  ## and eventually PM-MM
  if (pmonly){
    int <- log2(pmax(pmint, 1))
  }
  else {
    if (mmonly){
      int <- log2(pmax(mmint, 1))
    }
    else {
      int <- log2(pmax(pmint-mmint, 1))
    }
  }
  
  ## one-sample:
  ## simply take rowMedians
  ## over replicates
  if (is.null(custom2.filename)){
    if (ncol(int) > 1){
      data$intensity <- as.numeric(rowMedians(imat=int))
    }
    else {
      data$intensity <- as.numeric(int[,1])
    }
  }
  ## two samples:
  ## take paired median difference
  ## (median among all combinations
  ## of intensity differences in terms
  ## of custom2-custom)
  else {
    intcomb <- NULL
    for (i in 1:length(custom)){
      for (j in 1:length(custom2)){
        ## INFO: int stores custom intensities from
        ## 1:length(custom) and custom2 intensities from
        ## length(custom)+1:length(custom)+length(custom2)
        intcomb <- cbind(intcomb, int[,j+length(custom)]-int[,i])
      }
    }
    if (ncol(int) > 1){
      data$intensity <- as.numeric(rowMedians(imat=intcomb))
    }
    else {
      data$intensity <- as.numeric(intcomb[,1])
    }
    rm(intcomb)
  }

  ## order output by name, centerpos, and probelength
  data <- data[order(data$name, data$centerpos, data$probelength),]
  
  ## cleanup
  if (!mmonly){
    rm(pmint)
  }
  if (!pmonly){
    rm(mmint)
  }
  rm(custom, int)
  if (!is.null(custom2.filename)){
    rm(custom2)
  }
  if (!is.null(custominc.filename)){
    rm(custominc)
  }
  return(data)
}
